#include <conio.h>
#include <string.h>
#include <time.h>
#include "x_macros.h"
#include "x_crc32.hpp"
#include "x_err.hpp"
#include "port.hpp"
#include "unique.hpp"
#include "packet.hpp"

#ifdef COMPRESS
#include "compress.hpp"
#endif

#ifdef IPXEMU
#include "ipx.hpp"
#endif

#ifdef SETUP
#include "select.hpp"
#endif

void			findNodes();
#ifdef SETUP
void			doSetup();
#endif
void			sync();
int			getPacket();
int			readPacket(int *hdrP);
void			broadcastPacket();
void			sendPacket();
void			writePacket(int compress);
#ifdef DEBUG
void			packetDebug();
#endif
#ifdef HUFSTATS
void			hufStats();
#endif

static int		packetChecksum(int hdr, int pktLen, byte *pkt, int sending);
static int		readPacket(int *hdrP);
static void		sendCtrlPacket(int destNode, int ctrl, int dataLen = 0, void *data = NULL);

int			localStack;
byte			stack[2048];
word			savedFlags, savedSS, savedSP, savedBP, savedAX;
int			clientApproved;
int			nodeExists[MAX_NETNODES];
int			gameInProgress, initialNodes, nodeAddr;
int			broadcastReset, remoteAddr, maxAddr = -1;
int			packetLen;
byte			*packet;
byte			buf[(MAX_PACKET * 9 / 8 + 1) * 2 + 4];
int			bufLen;
word			nPacketsWaiting;
#ifdef SETUP
int			setupSource = MAX_NETNODES;
#endif
#ifdef DEBUG
word                    maxPacketsWaiting;
int			nMisc;
word			misc[64];
#endif

#ifdef DEBUG
static dword		minPacketLen = 0xffffffff, maxPacketLen;
static dword		nBadLen, maxBadLen;
static dword		nBroadcasts;
static dword		nSent, nRcvd;
static dword		nBadCksums;
static dword		nConnects, nDisconnects;
static dword		nSetupRequests, nSetupNotifications;
static dword		nResetRequests, nResetNotifications;
static dword            nSyncRequests, nSyncNotifications;
static dword            nDroppedSync, nDroppedControl;
static dword		pktLenHist[MAX_PACKET + 1];
#ifdef SAMPLE_PACKETS
static word		nSamplePackets;
static byte		samplePacket[SAMPLE_PACKET_N][SAMPLE_PACKET_LEN];
#endif
#endif
#ifdef HUFSTATS
static dword		pktByteHist[N_COMPRESSORS][256];
#endif
static int		awaitingSync;
static int		awaitSync[MAX_NETNODES];
#ifdef SETUP
static int		sentSetup[MAX_NETNODES];
#endif

#ifdef DEBUG
void packetDebug() {
  dword i, j, max1, max2;
  FILE *f = fopen("debug.txt", "at");
  fprintf(f, "--- PACKET Module Debug Info -----------------------------------\n");
  fprintf(f, "minPacketLen:         %lu\n", minPacketLen);
  fprintf(f, "maxPacketLen:         %lu\n", maxPacketLen);
  fprintf(f, "maxPacketsWaiting:    %u\n", maxPacketsWaiting);
  fprintf(f, "nBadLen:              %lu\n", nBadLen);
  fprintf(f, "maxBadLen:            %lu\n", maxBadLen);
  fprintf(f, "nBadCksums:           %lu\n", nBadCksums);
  fprintf(f, "nConnects:            %lu\n", nConnects);
  fprintf(f, "nDisconnects:         %lu\n", nDisconnects);
  fprintf(f, "nSetupRequests:       %lu\n", nSetupRequests);
  fprintf(f, "nSetupNotifications:  %lu\n", nSetupNotifications);
  fprintf(f, "nResetRequests:       %lu\n", nResetRequests);
  fprintf(f, "nResetNotifications:  %lu\n", nResetNotifications);
  fprintf(f, "nSyncRequests:        %lu\n", nSyncRequests);
  fprintf(f, "nSyncNotifications:   %lu\n", nSyncNotifications);
  fprintf(f, "nDroppedSync:         %lu\n", nDroppedSync);
  fprintf(f, "nDroppedControl:      %lu\n", nDroppedControl);
  for (i = 0; i < nMisc; i++)
    fprintf(f, "misc[%d]: %u (%x)\n", i, misc[i], misc[i]);
#ifdef IPXEMU
  fprintf(f, "nBroadcasts:          %lu\n", nBroadcasts);
#endif
  fprintf(f, "nSent:                %lu\n", nSent);
  fprintf(f, "nRcvd:                %lu\n", nRcvd);
#ifdef SAMPLE_PACKETS
  for (i = 0; i < SAMPLE_PACKET_LEN; i++) {
    fprintf(f, "%d: --------------------------------------------------------\n", i);
    for (int j = 0; j < nSamplePackets; j++) {
      fprintf(f, "%d\n", samplePacket[j][i]);
    }
  }
#endif
  for (max1 = 0, max2 = 0, j = 0, i = 0; i < MAX_PACKET; i++) {
    if (pktLenHist[i] > pktLenHist[max1]) {
      max2 = max1;
      max1 = i;
    } else if (pktLenHist[i] > pktLenHist[max2])
      max2 = i;
    j += pktLenHist[i];
  }
  if (j) {
    fprintf(f, "mode length:    %lu (%.2f%%)\n", max1, (float)pktLenHist[max1] / j * 100);
    fprintf(f, "mode length 2:  %lu (%.2f%%)\n", max2, (float)pktLenHist[max2] / j * 100);
  }
  fclose(f);
}
#endif

#ifdef HUFSTATS
void hufStats() {
  dword i, j, k;
  FILE *f = fopen("HUF.DAT", "wb");
  for (i = 0; i < N_COMPRESSORS; i++) {
    for (k = 0, j = 0; j < 256; j++) {
      if (!pktByteHist[i][j]) pktByteHist[i][j] = 1;
      if (pktByteHist[i][j] > k) k = pktByteHist[i][j];
    }
    if (k > 4096) {
      k = k / 255 + 1;
      for (j = 0; j < 256; j++) {
	pktByteHist[i][j] /= k;
	if (!pktByteHist[i][j]) pktByteHist[i][j] = 1;
      }
      for (j = 0; j < 256; j++) fputc(pktByteHist[i][j], f);
    }
  }
  fclose(f);
}
#endif

void findNodes() {
  int nodeExists[MAX_NETNODES];
  if (!gameInProgress) {
    for (int i = 0; i < initialNodes; i++)
      if (i != nodeAddr) {
	::nodeExists[i] = 1;
        maxAddr = max(maxAddr, i);
      }
    return;
  }
  printf("Scanning network");
  for (int i = 0; i < 15; i++) {
    for (int j = 0; j < 3; j++) sendCtrlPacket(63, 0);
    memcpy(nodeExists, ::nodeExists, MAX_NETNODES * sizeof(int));
    delay(1000);
    printf(".");
    while (nPacketsWaiting) getPacket();
    if (!memcmp(nodeExists, ::nodeExists, MAX_NETNODES * sizeof(int))) break;
  }
  printf("\n");
}

#ifdef SETUP
void doSetup() {
  if (!gameInProgress && !nodeAddr) {
    if (initialNodes == 1) return;
    printf("Sending game setup information");
    for (int t = 0; t < 30; t++) {
      int i;
      while (nPacketsWaiting) getPacket();
      for (i = 1; i <= maxAddr; i++)
        if (nodeExists[i] && !sentSetup[i]) break;
      if (i == maxAddr + 1) goto l0;
      delay(500);
      printf(".");
    }
  } else {
    printf("Receiving game setup information");
    while (kbhit()) getch();
    for (;;) {
      if (kbhit()) {
        while (kbhit()) getch();
	fatalError("Setup aborted");
      }
      if (gameInProgress) sendCtrlPacket(nodeAddr, CTRL_SETUP_REQUEST);
      else sendCtrlPacket(0, CTRL_SETUP_REQUEST);
      delay(500);
      printf(".");
      while (nPacketsWaiting) getPacket();
      if (gameInProgress && setupSource < MAX_NETNODES) goto l0;
      else if (!setupSource) goto l0;
    }
  }
  fatalError("Setup failed");
l0:
  setupKnown = 1;
  printf("\n");
}
#endif

int packetChecksum(int hdr, int pktLen, byte *pkt, int sending) {
  int i;
  static CRC32 crc32;
  crc32.init();
  if (sending) crc32.addByte(DEST_HDR(hdr));
  else crc32.addByte(hdr);
  for (i = 0; i < pktLen; i++)
    crc32.addByte(pkt[i]);
  return crc32.CRC() & 0x7fff;
}

void sync() {
  for (int i = 0; i <= maxAddr; i++)
    if (nodeExists[i]) {
      awaitingSync = 1;
      for (int j = 0; j < 3; j++) sendCtrlPacket(i, CTRL_SYNC_REQUEST);
      awaitSync[i] = 1;
    }
}

int getPacket() {
  int i, hdr;
l0:
  if (!nPacketsWaiting) return 1;
  if (!readPacket(&hdr)) {
    if (PACKET_CONTROL(hdr)) {
      if (!packetLen) goto l0;
      switch (buf[0]) {
      case CTRL_CONNECT:
	if (packetLen != 2 || buf[1] >= MAX_NETNODES) goto l0;
#ifdef DEBUG
	nConnects++;
#endif
	nodeExists[buf[1]] = 1;
	maxAddr = max(maxAddr, buf[1]);
	break;
      case CTRL_DISCONNECT:
	if (packetLen != 2 || buf[1] >= MAX_NETNODES) goto l0;
#ifdef DEBUG
	nDisconnects++;
#endif
#ifdef COMPRESS
	compressInit(buf[1], CTRL_RESET_REQUEST);
	compressInit(buf[1], CTRL_RESET_NOTIFY);
#endif
	nodeExists[buf[1]] = 0;
	if (maxAddr == buf[1])
	  for (maxAddr--; maxAddr >= 0; maxAddr--)
	    if (nodeExists[maxAddr]) break;
	break;
#ifdef SETUP
      case CTRL_SETUP_REQUEST:
#ifdef DEBUG
	nSetupRequests++;
#endif
        if (setupKnown) {
          sentSetup[remoteAddr] = 1;
          sendCtrlPacket(remoteAddr, CTRL_SETUP_NOTIFY, sizeof(Setup), &setup);
	}
	break;
      case CTRL_SETUP_NOTIFY:
#ifdef DEBUG
	nSetupNotifications++;
#endif
	if (packetLen != sizeof(Setup) + 1) goto l0;
        memcpy(&setup, buf + 1, sizeof(Setup));
	setupSource = min(setupSource, remoteAddr);
	break;
#endif
#ifdef COMPRESS
      case CTRL_RESET_REQUEST:
#ifdef DEBUG
	nResetRequests++;
#endif
	compressInit(remoteAddr, CTRL_RESET_REQUEST);
	sendCtrlPacket(remoteAddr, CTRL_RESET_NOTIFY);
	broadcastReset = 1;
	break;
      case CTRL_RESET_NOTIFY:
#ifdef DEBUG
	nResetNotifications++;
#endif
	compressInit(remoteAddr, CTRL_RESET_NOTIFY);
	break;
#endif
      case CTRL_SYNC_REQUEST:
#ifdef DEBUG
	nSyncRequests++;
#endif
	sendCtrlPacket(remoteAddr, CTRL_SYNC_NOTIFY);
	break;
      case CTRL_SYNC_NOTIFY:
#ifdef DEBUG
	nSyncNotifications++;
#endif
	awaitSync[remoteAddr] = 0;
	for (i = 0; i <= maxAddr; i++)
	  if (awaitSync[i]) break;
	if (i == maxAddr + 1) awaitingSync = 0;
	break;
#ifdef DEBUG
      default:
	nDroppedControl++;
#endif
      }
      goto l0;
    } else return 0;
  } else goto l0;
}

int readPacket(int *hdrP) {
  int err, control, compressed, pktLen = 0, pktHdr, pktCksum, esc = 0;
  byte c, *pkt;
  portRead(&c);
  *hdrP = pktHdr = c;
  control = PACKET_CONTROL(pktHdr);
  compressed = PACKET_COMPRESSED(pktHdr);
  remoteAddr = PACKET_ADDR(pktHdr);
  pkt = buf;
  portRead(&c);
  pktCksum = c;
  for (;;) {
    portRead(&c);
    if (esc) {
      esc = 0;
      if (c != FRAME_CHAR) {
        nPacketsWaiting--;
	if (awaitingSync && !control) {
#ifdef DEBUG
          nDroppedSync++;
#endif
          return 1;
        }
	if (remoteAddr >= MAX_NETNODES ||
	    (compressed && control) ||
            (!control && !nodeExists[remoteAddr]))
	{
#ifdef DEBUG
          nDroppedControl++;
#endif
	  return 1;
	}
        pktCksum |= c << 8;
#ifdef COMPRESS
	if (compressed) {
	  err = uncompressPacket(pkt);
	  if (!err) err = packetChecksum(pktHdr, packetLen, packet, 0) != pktCksum;
	  if (err) sendCtrlPacket(remoteAddr, CTRL_RESET_REQUEST);
	} else
#endif
        {
          err = packetChecksum(pktHdr, pktLen, pkt, 0) != pktCksum;
	  if (!err) {
	    packetLen = pktLen;
	    if (!control) memcpy(packet, pkt, packetLen);
	  }
	}
	if (err) {
#ifdef DEBUG
	  nBadCksums++;
#endif
          return 1;
	} else {
#ifdef DEBUG
          nRcvd++;
#endif
          return 0;
	}
      }
    } else if (c == FRAME_CHAR) {
      esc = 1;
      continue;
    }
    if (pktLen == MAX_PACKET) continue;
    pkt[pktLen++] = c;
  }
}

void broadcastPacket() {
  remoteAddr = nodeAddr;
  sendPacket();
#ifdef DEBUG
  nBroadcasts++;
#endif
}

void sendPacket() {
  int i, cksum;
  byte pktHdr;
  if (awaitingSync) return;
  if (packetLen < 0 || packetLen > MAX_PACKET) {
#ifdef DEBUG
    nBadLen++;
    if (packetLen > maxBadLen) maxBadLen = packetLen;
#endif
    return;
  }
#ifdef DEBUG
  pktLenHist[packetLen]++;
  if (packetLen < minPacketLen) minPacketLen = packetLen;
  else if (packetLen > maxPacketLen) maxPacketLen = packetLen;
#endif
#ifdef SAMPLE_PACKETS
  if (packetLen == SAMPLE_PACKET_LEN && nSent > SAMPLE_PACKET_FIRST && nSamplePackets < SAMPLE_PACKET_N)
    _fmemcpy(samplePacket[nSamplePackets++], packet, packetLen);
#endif
#ifdef HUFSTATS
  if (COMPRESSABLE(packetLen))
    for (int i = 0; i < packetLen; i++)
      if (cMethod[i] >= 0)
	pktByteHist[cMethod[i]][packet[i]]++;
#endif
  pktHdr = 0;
#ifdef COMPRESS
  if (COMPRESSABLE(packetLen)) {
    if (remoteAddr == nodeAddr) {
      if (broadcastReset) {
	compressInit(nodeAddr, CTRL_RESET_REQUEST);
        sendCtrlPacket(nodeAddr, CTRL_RESET_NOTIFY);
        broadcastReset = 0;
      }
    } else broadcastReset = 1;
    compressPacket();
    pktHdr |= PKT_COMPRESSED;
  }
#endif
  pktHdr |= remoteAddr;
  cksum = packetChecksum(pktHdr, packetLen, packet, 1);
  bufLen = 0;
  buf[bufLen++] = pktHdr;
  buf[bufLen++] = cksum & 0xff;
#ifdef COMPRESS
  if (pktHdr & PKT_COMPRESSED) {
    for (i = 0; i < cPacketLen; i++) {
      if (cPacket[i] == FRAME_CHAR) buf[bufLen++] = FRAME_CHAR;
      buf[bufLen++] = cPacket[i];
    }
  } else
#endif
  for (i = 0; i < packetLen; i++) {
    if (packet[i] == FRAME_CHAR) buf[bufLen++] = FRAME_CHAR;
    buf[bufLen++] = packet[i];
  }
  buf[bufLen++] = FRAME_CHAR;
  buf[bufLen++] = (cksum & 0x7f00) >> 8;
  if (!portWrite(buf, bufLen)) {
#ifdef DEBUG
    nSent++;
#endif
  }
}

void sendCtrlPacket(int destNode, int ctrl, int dataLen, void *data) {
  int hdr = PKT_CONTROL | destNode;
  int pktLen = 1;
  buf[0] = ctrl;
  if (dataLen) {
    memcpy(buf + 1, data, dataLen);
    pktLen += dataLen;
  }
  int cksum = packetChecksum(hdr, pktLen, buf, 1);
  portWrite(hdr);
  portWrite(cksum & 0xff);
  portWrite(buf, pktLen);
  portWrite(FRAME_CHAR);
  portWrite((cksum & 0x7f00) >> 8);
}
