
#include <ctype.h>
#include <dirent.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include "ansi.h"
#include "lgc.h"
#include "output.h"
#include "help.h"

const int                MAX_HELPTOPIC_LEN = 16;

int
help(char *arg)
{
  int                      c,
                           i,
                           ti,
                           x,
                           y,
                           res = 0,
                           len = 0,
                           interpret = 0;
  char                     token[MAX_LINELEN],
                           context[PATH_MAX + 1],
                           topic[MAX_HELPTOPIC_LEN + 1];
  fd_set                   rSet,
                           allSet;

  context[0] = '\0';
  topic[0] = '\0';
  FD_ZERO(&allSet);
  FD_SET(STDIN_FILENO, &allSet);
  FD_SET(fd, &allSet);
  if (!arg)
    helpListTopics(context);
  else
    helpListTopic(context, arg);
  if (optDOS) {
    DOSwriteByte(PKT_PROMPT);
    DOSwriteString("HELP> ", 6);
    DOSendPacket();
  } else {
    _werase(cmdW);
    waddstr(cmdW, "HELP> ");
    wrefresh(cmdW);
  }
  for (;;) {
    rSet = allSet;
    select(fd + 1, &rSet, NULL, NULL, NULL);
    if (FD_ISSET(fd, &rSet))
      while (!clientRead());
    while (nMsgsWaiting) {
      getMessage(&msg);
      processMessage();
    }
    if (FD_ISSET(STDIN_FILENO, &rSet)) {
      if (optDOS)
	c = getchar();
      else
	c = wgetch(cmdW);
      if (isprint(c)) {
	if (len < MAX_HELPTOPIC_LEN) {
	  topic[len++] = c;
	  if (!optDOS)
	    waddch(cmdW, c);
	} else if (!optDOS)
	  beep();
      } else if (c == 1) {
	if (optDOS) {
	  DOSwriteByte(PKT_PROMPT);
	  DOSwriteString("HELP", 4);
	} else {
	  _werase(cmdW);
	  waddstr(cmdW, "HELP");
	}
	if (context[0]) {
	  if (optDOS) {
	    DOSwriteByte(':');
	    DOSwriteString(context, strlen(context));
	  } else {
	    waddstr(cmdW, ":");
	    waddstr(cmdW, context);
	  }
	}
	if (optDOS) {
	  DOSwriteString("> ", 2);
	  DOSendPacket();
	} else
	  waddstr(cmdW, "> ");
	len = 0;
      } else if (c == 8 || c == 127 || c == 263) {
	if (len) {
	  if (!optDOS) {
	    getyx(cmdW, y, x);
	    wmove(cmdW, y, x - 1);
	    waddch(cmdW, ' ');
	    wmove(cmdW, y, x - 1);
	    wrefresh(cmdW);
	  }
	  len--;
	} else if (!optDOS)
	  beep();
      } else if (c == 10 || c == 13) {
	interpret = 1;
      } else if (c == 12) {
	if (!optDOS) {
	  clearok(curscr, TRUE);
	  refresh();
	}
      } else if (isspace(c)) {
	topic[len++] = ' ';
	if (!optDOS)
	  waddch(cmdW, ' ');
      } else if (!optDOS)
	beep();
      if (interpret) {
	interpret = 0;
	if (!optDOS)
	  _werase(cmdW);
	topic[len] = '\0';
	ti = 0;
	if (!getToken(token, topic, &ti)) {
	  topic[len] = '\0';
	  if (!strncasecmp(token, "/list", 2))
	    helpListTopics(context);
	  else if (!strncasecmp(token, "/exit", 2))
	    break;
	  else if (!strncasecmp(token, "/quit", 2)) {
	    res = 1;
	    break;
	  } else
	    helpListTopic(context, topic);
	} else {
	  i = strlen(context) - 1;
	  if (i >= 0) {
	    while (i && context[i] != '/')
	      i--;
	    context[i] = '\0';
	    helpListTopics(context);
	  } else
	    break;
	}
	if (optDOS) {
	  DOSwriteByte(PKT_PROMPT);
	  DOSwriteString("HELP", 4);
	} else
	  waddstr(cmdW, "HELP");
	if (context[0]) {
	  if (optDOS) {
	    DOSwriteByte(':');
	    DOSwriteString(context, strlen(context));
	  } else {
	    waddstr(cmdW, ":");
	    waddstr(cmdW, context);
	  }
	}
	if (optDOS) {
	  DOSwriteString("> ", 2);
	  DOSendPacket();
	} else
	  waddstr(cmdW, "> ");
	len = 0;
      }
      if (!optDOS)
	wrefresh(cmdW);
    }
  }
  if (optDOS) {
    DOSwriteByte(PKT_PROMPT);
    DOSendPacket();
  }
  return res;
}

void
helpListTopics(char *context)
{
  struct stat              _stat;
  struct dirent           *_dirent;
  int                      i,
                           len,
                           filesonly,
                           n_printed = 0,
                           max_tpl = COLS / (MAX_HELPTOPIC_LEN + 4);
  char                     path[PATH_MAX + 1],
                           tpath[PATH_MAX + 1];
  DIR                     *dir;

  msgStr("\n<B>Help Listing</B>:  (all commands preceded by '/')\n\n");
  sprintf(path, "%s/help%s", LIB, context);
  for (filesonly = 0; filesonly <= 1; filesonly++) {
    dir = opendir(path);
    while ((_dirent = readdir(dir)) != NULL) {
      if (strcmp(_dirent->d_name, ".") && strcmp(_dirent->d_name, "..")) {
	sprintf(tpath, "%s/%s", path, _dirent->d_name);
	lstat(tpath, &_stat);
	if (S_ISDIR(_stat.st_mode) && !filesonly) {
	  len = strlen(_dirent->d_name);
	  for (i = 0; i < len; i++)
	    _dirent->d_name[i] = toupper(_dirent->d_name[i]);
	  msgStr("<B>%-19s</B>", _dirent->d_name);
	  n_printed++;
	  if (!(n_printed % max_tpl))
	    msgStr("\n");
	} else if (!S_ISDIR(_stat.st_mode) && filesonly) {
	  len = strlen(_dirent->d_name);
	  for (i = 0; i < len; i++)
	    _dirent->d_name[i] = toupper(_dirent->d_name[i]);
	  msgStr("%-19s", _dirent->d_name);
	  n_printed++;
	  if (!(n_printed % max_tpl))
	    msgStr("\n");
	}
      }
    }
    closedir(dir);
  }
  if (n_printed % max_tpl)
    msgStr("\n");
  msgStr("\n");
}

void
helpListTopic(char *context, char *topic)
{
  int                      i,
                           len,
                           found = 0;
  struct stat              _stat;
  struct dirent           *_dirent;
  FILE                    *f;
  char                     path[PATH_MAX + 1],
                           tpath[PATH_MAX + 1],
                           line[256];
  DIR                     *dir;

  sprintf(path, "%s/help%s", LIB, context);
  dir = opendir(path);
  while ((_dirent = readdir(dir)) != NULL) {
    if (strcmp(_dirent->d_name, ".") && strcmp(_dirent->d_name, "..")) {
      sprintf(tpath, "%s/%s", path, _dirent->d_name);
      lstat(tpath, &_stat);
      if (!strncasecmp(_dirent->d_name, topic, strlen(topic))) {
	found = 1;
	if (S_ISDIR(_stat.st_mode)) {
	  strcat(context, "/");
	  strcat(context, _dirent->d_name);
	  helpListTopics(context);
	} else {
	  len = strlen(_dirent->d_name);
	  for (i = 0; i < len; i++)
	    _dirent->d_name[i] = toupper(_dirent->d_name[i]);
	  msgStr("\n");
	  paging(1);
	  f = fopen(tpath, "rt");
	  msgStr("<C %d>**</C> Help on %s\n", FG_STAR, _dirent->d_name);
	  while (fgets(line, 256, f) != NULL)
	    msgStr(line);
	  fclose(f);
	  msgStr("\n");
	  paging(0);
	  break;
	}
      }
    }
  }
  closedir(dir);
  if (!found)
    msgStr("<C %d>**</C> No match for '%s'\n", FG_STAR, topic);
}
